/* Langton's Ant
 * Author: Aaron Weeden, Shodor, 2015
 */
#include <mpi.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include <X11/Xlib.h>

// Declare constants
#define TIME_STEP_COUNT 200
#define ANT_COUNT 10
#define PX_PER_CELL_X 5
#define PX_PER_CELL_Y 5
#define PX_PER_ANT_X 2
#define PX_PER_ANT_Y 2
#define CELL_COUNT_X 21
#define CELL_COUNT_Y 21
#define CELL_COUNT ((CELL_COUNT_X) * (CELL_COUNT_Y))
#define MICROS_PER_SLEEP 5000
#define RANDOM_SEED 0
#define PX_PER_GRID_X ((PX_PER_CELL_X) * (CELL_COUNT_X))
#define PX_PER_GRID_Y ((PX_PER_CELL_Y) * (CELL_COUNT_Y))

#define BLACK 0
#define WHITE 1

#define UP    0
#define LEFT  1
#define RIGHT 2
#define DOWN  3

// Map 2D coordinates to 1D coordinates
#define IDX(x, y) ((x) * (CELL_COUNT_X) + (y))

void divvyUp(int const total, int const mpiSize, int *counts, int *displs) {
  int const quotient  = total / mpiSize;
  int const remainder = total % mpiSize;
  for (int i = 0; i < mpiSize; i++) {
    if (i < remainder) {
      counts[i] = quotient + 1;
      displs[i] = i * (quotient + 1);
    }
    else {
      counts[i] = quotient;
      displs[i] = remainder * (quotient + 1) + (i - remainder) * quotient;
    }
  }
}

void drawAndPause(Display * const x11Display, Window const x11Window,
                  GC const x11GC, XColor const x11WhiteColor,
                  XColor const x11BlackColor, XColor const x11RedColor,
                  int const timeStep,
                  int cellStates[CELL_COUNT],
                  int const * antDirs, int const * antCellXs,
                  int const * antCellYs) {
  printf("Time step %d\n", timeStep);

  // Draw cells
  XClearWindow(x11Display, x11Window);
  for (int x = 0; x < CELL_COUNT_X; x++) {
    for (int y = 0; y < CELL_COUNT_Y; y++) {
      XSetForeground(x11Display,
                     x11GC,
                     (cellStates[IDX(x, y)] == BLACK) ? x11BlackColor.pixel :
                                                        x11WhiteColor.pixel);

      // Print cell color
      printf("%c", cellStates[IDX(x, y)] == BLACK ? 'B' : '_');
      XFillRectangle(x11Display,
                     x11Window,
                     x11GC,
                     x * PX_PER_CELL_X,
                     y * PX_PER_CELL_Y,
                     PX_PER_CELL_X,
                     PX_PER_CELL_Y);

      // Draw ant
      XSetForeground(x11Display, x11GC, x11RedColor.pixel);
      bool isAntHere = false;
      for (int k = 0; k < ANT_COUNT; k++) {
        if (antCellXs[k] == x && antCellYs[k] == y) {
          int antPxX;
          int antPxY;
          switch (antDirs[k]) {
            case UP:
              printf("U");
              antPxX = antCellXs[k] * PX_PER_CELL_X +
                       0.5 * (PX_PER_CELL_X - PX_PER_ANT_X);
              antPxY = antCellYs[k] * PX_PER_CELL_Y;
              break;
            case LEFT:
              printf("L");
              antPxX = antCellXs[k] * PX_PER_CELL_X;
              antPxY = antCellYs[k] * PX_PER_CELL_Y +
                       0.5 * (PX_PER_CELL_Y - PX_PER_ANT_Y);
              break;
            case RIGHT:
              printf("R");
              antPxX = (antCellXs[k] + 1) * PX_PER_CELL_X - PX_PER_ANT_X;
              antPxY = antCellYs[k] * PX_PER_CELL_Y +
                       0.5 * (PX_PER_CELL_Y - PX_PER_ANT_Y);
              break;
            case DOWN:
              printf("D");
              antPxX = antCellXs[k] * PX_PER_CELL_X +
                       0.5 * (PX_PER_CELL_X - PX_PER_ANT_X);
              antPxY = (antCellYs[k] + 1) * PX_PER_CELL_Y - PX_PER_ANT_Y;
              break;
          }
          XFillRectangle(x11Display, x11Window, x11GC, antPxX, antPxY,
                         PX_PER_ANT_X, PX_PER_ANT_Y);
          // Only print direction for one ant
          isAntHere = true;
          break;
        }
      }
      if (!isAntHere) {
        printf(" ");
      }
      printf(" ");
    }
    printf("\n");
  }
  printf("\n");
  XFlush(x11Display);

  // Pause
  usleep(MICROS_PER_SLEEP);
}

int main(int argc, char **argv) {
  // Initialize MPI
  MPI_Init(&argc, &argv);
  int myRank;
  MPI_Comm_rank(MPI_COMM_WORLD, &myRank);
  int mpiSize;
  MPI_Comm_size(MPI_COMM_WORLD, &mpiSize);

  // Initialize the X Windows Environment
  // This all comes from
  //   http://en.wikibooks.org/wiki/X_Window_Programming/XLib
  //   http://tronche.com/gui/x/xlib-tutorial
  //   http://user.xmission.com/~georgeps/documentation/tutorials/
  //      Xlib_Beginner.html
  Display *x11Display;
  Window x11Window;
  GC x11GC;
  XColor x11WhiteColor;
  XColor x11BlackColor;
  XColor x11RedColor;
  int err = false;
  if (myRank == 0) {
    x11Display = XOpenDisplay(NULL);
    if(x11Display == NULL) {
      err = true;
    }
    int const x11Screen = DefaultScreen(x11Display);
    x11Window = XCreateSimpleWindow(x11Display,
                                    RootWindow(x11Display, x11Screen),
                                    0,
                                    0,
                                    PX_PER_CELL_X  * CELL_COUNT_X,
                                    PX_PER_CELL_Y  * CELL_COUNT_Y,
                                    1,
                                    BlackPixel(x11Display, x11Screen),
                                    WhitePixel(x11Display, x11Screen));
    Atom deleteWindow = XInternAtom(x11Display, "WM_DELETE_WINDOW", 0);
    XSetWMProtocols(x11Display, x11Window, &deleteWindow, 1);
    XSelectInput(x11Display, x11Window, ExposureMask | KeyPressMask);
    XMapWindow(x11Display, x11Window);
    Colormap const x11ColorMap = DefaultColormap(x11Display, 0);
    x11GC = XCreateGC(x11Display, x11Window, 0, 0);
    XParseColor(x11Display, x11ColorMap, "#000000", &x11BlackColor);
    XParseColor(x11Display, x11ColorMap, "#FFFFFF", &x11WhiteColor);
    XParseColor(x11Display, x11ColorMap, "#FF0000", &x11RedColor);
    XAllocColor(x11Display, x11ColorMap, &x11BlackColor);
    XAllocColor(x11Display, x11ColorMap, &x11WhiteColor);
    XAllocColor(x11Display, x11ColorMap, &x11RedColor);
  }
  int err2;
  MPI_Allreduce(&err, &err2, 1, MPI_INT, MPI_LOR, MPI_COMM_WORLD);
  if (err2) {
    fprintf(stderr, "ERROR: could not open X display\n");
    MPI_Finalize();
    exit(EXIT_FAILURE);
  }

  // Initialize cell states
  int cellStates[CELL_COUNT];
  for (int i = 0; i < CELL_COUNT; i++) {
    cellStates[i] = WHITE;
  }

  // For each ant
  int antCellXs[ANT_COUNT];
  int antCellYs[ANT_COUNT];
  int antDirs[ANT_COUNT];
  srandom(RANDOM_SEED);
  for (int i = 0; i < ANT_COUNT; i++) {
    // Give the ant a position
    if (myRank == 0) {
      antCellXs[i] = random() % CELL_COUNT_X;
      antCellYs[i] = random() % CELL_COUNT_Y;
    }

    // Give the ant a direction
    antDirs[i] = LEFT;
  }
  // Synchronize ant positions
  MPI_Bcast(antCellXs, ANT_COUNT, MPI_INT, 0, MPI_COMM_WORLD);
  MPI_Bcast(antCellYs, ANT_COUNT, MPI_INT, 0, MPI_COMM_WORLD);

  if (myRank == 0) {
    drawAndPause(x11Display, x11Window, x11GC, x11WhiteColor, x11BlackColor,
                 x11RedColor, 0, cellStates, antDirs, antCellXs, antCellYs);
  }

  // Divvy up cells among processes
  int cellCounts[mpiSize];
  int cellDispls[mpiSize];
  divvyUp(CELL_COUNT, mpiSize, cellCounts, cellDispls);

  // Divvy up ants among processes
  int antCounts[mpiSize];
  int antDispls[mpiSize];
  divvyUp(ANT_COUNT, mpiSize, antCounts, antDispls);

  // Start the simulation
  for (int timeStep = 0; timeStep < TIME_STEP_COUNT; timeStep++) {
    // Turn
    int myAntDirs[antCounts[myRank]];
    for (int i = 0; i < antCounts[myRank]; i++) {
      int const oldI = i;
      i += antDispls[myRank];
      bool const isBlack = (cellStates[IDX(antCellXs[i],
                                           antCellYs[i])] == BLACK);
      if (     (antDirs[i] == UP    &&  isBlack) ||
               (antDirs[i] == DOWN  && !isBlack)) {
        myAntDirs[oldI] = LEFT;
      }
      else if ((antDirs[i] == UP    && !isBlack) ||
               (antDirs[i] == DOWN  &&  isBlack)) {
        myAntDirs[oldI] = RIGHT;
      }
      else if ((antDirs[i] == LEFT  &&  isBlack) ||
               (antDirs[i] == RIGHT && !isBlack)) {
        myAntDirs[oldI] = DOWN;
      }
      else if ((antDirs[i] == LEFT  && !isBlack) ||
               (antDirs[i] == RIGHT &&  isBlack)) {
        myAntDirs[oldI] = UP;
      }
      i = oldI;
    }

    // Synchronize ant directions
    MPI_Allgatherv(myAntDirs, antCounts[myRank], MPI_INT, antDirs, antCounts,
                   antDispls, MPI_INT, MPI_COMM_WORLD);

    if (myRank == 0) {
      drawAndPause(x11Display, x11Window, x11GC, x11WhiteColor, x11BlackColor,
                   x11RedColor, timeStep, cellStates, antDirs, antCellXs,
                   antCellYs);
    }

    // Update cells
    int shouldFlipCells[CELL_COUNT];
    int myShouldFlipCells[CELL_COUNT];
    for (int i = 0; i < CELL_COUNT; i++) {
      myShouldFlipCells[i] = false;
    }
    for (int i = antDispls[myRank];
             i < antDispls[myRank] + antCounts[myRank];
             i++) {
      myShouldFlipCells[IDX(antCellXs[i], antCellYs[i])] = true;
    }

    // Synchronize the "should flip cells" list
    MPI_Allreduce(myShouldFlipCells, shouldFlipCells, CELL_COUNT, MPI_INT,
                  MPI_LOR, MPI_COMM_WORLD);

    for (int i = 0; i < CELL_COUNT; i++) {
      if (shouldFlipCells[i]) {
        cellStates[i] = (cellStates[i] == BLACK) ? WHITE : BLACK;
      }
    }

    if (myRank == 0) {
      drawAndPause(x11Display, x11Window, x11GC, x11WhiteColor, x11BlackColor,
                   x11RedColor, timeStep, cellStates, antDirs, antCellXs,
                   antCellYs);
    }

    // Move
    int myAntCellXs[antCounts[myRank]];
    int myAntCellYs[antCounts[myRank]];
    for (int i = 0; i < antCounts[myRank]; i++) {
      int const oldI = i;
      i += antDispls[myRank];
      myAntCellXs[oldI] = antCellXs[i];
      myAntCellYs[oldI] = antCellYs[i];
      switch (antDirs[i]) {
        case UP:
          if (antCellYs[i] == 0) {
            myAntCellYs[oldI] = CELL_COUNT_Y-1;
          }
          else {
            myAntCellYs[oldI]--;
          }
          break;
        case LEFT:
          if (antCellXs[i] == 0) {
            myAntCellXs[oldI] = CELL_COUNT_X-1;
          }
          else {
            myAntCellXs[oldI]--;
          }
          break;
        case RIGHT:
          if (antCellXs[i] == CELL_COUNT_X-1) {
            myAntCellXs[oldI] = 0;
          }
          else {
            myAntCellXs[oldI]++;
          }
          break;
        case DOWN:
          if (antCellYs[i] == CELL_COUNT_Y-1) {
            myAntCellYs[oldI] = 0;
          }
          else {
            myAntCellYs[oldI]++;
          }
          break;
      }
      i = oldI;
    }

    // Synchronize ant positions
    MPI_Allgatherv(myAntCellXs, antCounts[myRank], MPI_INT, antCellXs,
                   antCounts, antDispls, MPI_INT, MPI_COMM_WORLD);
    MPI_Allgatherv(myAntCellYs, antCounts[myRank], MPI_INT, antCellYs,
                   antCounts, antDispls, MPI_INT, MPI_COMM_WORLD);

    if (myRank == 0) {
      drawAndPause(x11Display, x11Window, x11GC, x11WhiteColor, x11BlackColor,
                   x11RedColor, timeStep, cellStates, antDirs, antCellXs,
                   antCellYs);
    }
  } // Continue animation loop

  MPI_Finalize();

  return 0;
}

